/**
 * Project:  Nano Waveform Generator  
 * ````````
 * File:     nano-wave-gen-main.c
 * `````
 * Firmware for the AVR "Nano" waveform generator
 * ``````````````````````````````````````````````
 * Reference:  "Build the Nano Waveform Generator" by M.J.Bauer 2022
 * ``````````` 
 * Originated:  M.J.Bauer, April 2021  [www.mjbauer.biz]
 * ```````````
 *
 * Brief specification:
 * ````````````````````
 * Modes:   Wave  - sine, triangle, square, sawtooth (+|- slope)
 *          Noise - variable cut-off freq. (IIR filter)
 *          Pulse - variable duty
 * Frequency:  Wave (Low range):   1 Hz ~ 100 Hz (12 preset steps)
 *             Wave (High range):  80 Hz ~ 8 kHz (18 preset steps)
 *                            or:  50 Hz ~ 5 kHz (variable using pot)
 *             Pulse (Low range):  1 Hz ~ 1000 Hz (16 preset steps)
 *             Pulse (High range): 1 kHz ~ 4 MHz (16 preset steps)
 * Duty:       1 ~ 99 % (variable using pot, quantized)
 * Output:  Wave/Noise o/p Level:  0.1, 0.2, 0.5, 1, 2, 4 Vp-p 
 *          Pulse: 3.3V or 5V fixed level, 20mA source/sink (5V)
 * PWM DAC resolution:  8 bits (0.4% FS, approx 4mV at 1Vpp)
 * 
 * Operating Instructions:
 * ```````````````````````
 * To select output mode (waveform/noise/pulse), press MODE/RANGE button.
 * To set frequency, press FREQ (+)|(-) button -OR- rotate the FREQ/DUTY knob.
 * To toggle Hi/Low range: Press (hold) ALT button and press MODE/RANGE button.
 * To set output level: Press (hold) ALT button and press LEVEL (+)|(-) button.
 * To set pulse duty (% or Tpw) in PULSE mode, rotate the FREQ/DUTY knob.
 * To view firmware version, Press (hold) ALT button during power-up.
 */
#include <avr/io.h>
#define F_CPU  16000000UL    // Arduino Nano CPU runs at 16 MHz

#include <avr/interrupt.h>  
#include <util/delay.h>   
#include <stdlib.h>
#include <string.h>
#include <avr/pgmspace.h>    // Functions to manage flash PM data
#include "lib_avrXmini.h"    // AVR peripheral code library

//================================//
#define FIRMWARE_VERSION   "1.5"  //
//================================//

// Macros to control Timer-Counter TC0 
#define TC0_OCA_IRQ_ENABLE()    (TIMSK0 |= (1<<OCIE0A))
#define TC0_OCA_IRQ_DISABLE()   (TIMSK0 &= ~(1<<OCIE0A))
#define TC0_WRITE_PRESCALER(n)  (TCCR0B = (TCCR0B & 0xF8) | (n & 7))

#define ALT_BUTTON_IS_PRESSED   ((PIND & (1<<7)) == 0)

#define WAVE_TABLE_SIZE      ((unsigned long) 640)
#define SAMPLE_RATE          32000

#define MODE_SINE            0
#define MODE_TRIANGLE        1
#define MODE_SQUARE          2
#define MODE_SAWTOOTH_POS    3
#define MODE_SAWTOOTH_NEG    4
#define MODE_PULSE           5
#define MODE_NOISE           6
#define MODE_UNDEFINED       7

#define LOW_RANGE                0
#define HIGH_RANGE               1
#define SET_FREQ_USING_BUTTONS  'B'
#define SET_FREQ_USING_POT      'P'
#define STEP_UP                  1
#define NO_CHANGE                0
#define STEP_DOWN                0xFF  // (-1)

// Define 32-bit fixed-point data type with 16-bit integer part, 16-bit fraction
typedef unsigned long  fixed_point_t; 

// ------------  Function prototype declarations  ----------------------------
void  TC0_Setup(BYTE wgm, BYTE prescale, BYTE period, BYTE duty);
void  ConfigureOutputMode();
void  SetOscFreqUsingPot();
void  SetNoiseFreqUsingPot(); 
void  SetPulseDutyUsingPot();
void  AdjustOscFreqPreset(BYTE dirn);
void  AdjustPulseFreqPreset(BYTE dirn);
void  AdjustOutputLevel(BYTE dirn);
void  DisplayUpdate();
void  TC1_TimerInit();
BOOL  TimerExpired_50ms();
void  ReverseOrder(BYTE *dest, unsigned size);
void  lcd_erase_field(BYTE row, BYTE col, BYTE nplaces);
void  PGM_ReadData(const BYTE *src, BYTE *dest, unsigned nbytes);

// ------------  Global variables and data  ----------------------------------
BYTE   waveTable[640];             // Active wave-table (RAM buffer)
BYTE   startupTimer;               // Timer-counter, unit = 50ms
BYTE   outputMode;                 // Output mode -> waveform
BYTE   freqSettingMode;            // Freq. setting mode, 'Buttons' or 'Pot'
BYTE   oscFreqRange;               // Wave Osc freq range, 0:Low, 1:High
BYTE   freqPresetLow;              // Wave Osc freq preset, Low range
BYTE   freqPresetHigh;             // Wave Osc freq preset, High range
BYTE   pulseFreqRange;             // Pulse freq range, 0:Low, 1:High
BYTE   pulsePresetLow;             // Pulse freq preset, Low range
BYTE   pulsePresetHigh;            // Pulse freq preset, High range
BYTE   pulseDuty_pc;               // Pulse duty, percent (for display)
BYTE   noiseFilterIdx;             // Noise filter preset (idx = log2(1/K))
BYTE   outputLevel;                // Output level setting (0..5)
BYTE   potSetting;                 // Pot setting, 0..255 (current position)

unsigned long  oscFreq_Hz;         // Osc. freq. (1 ~ 5000 Hz)
unsigned long  pulseFreq_Hz;       // Pulse freq. (1 Hz ~ 2 MHz)
unsigned long  pulseWidth_ns;      // Pulse width, unit = ns (for display)
fixed_point_t  oscStep;            // Osc. phase angle step, fixed-point
fixed_point_t  pulseDuty_fxt;      // Pulse duty, fixed-point

// Wave-tables defined in separate file: "nano-wave-data.c"
extern const BYTE sineWave[] PROGMEM; 
extern const BYTE triangleWave[] PROGMEM; 
extern const BYTE squareWave[] PROGMEM; 
extern const BYTE sawtoothWave[] PROGMEM;         

//============================================================================

int  main(void)
{
    static BYTE  lastPotSetting;  
    static BYTE  count_to_4;
    
    // Configure I/O pins for the application...
    SET_BIT(DDRB, 5);             // PB5 is pulse mode 'LRANGE' output
    SET_BIT(DDRB, 4);             // PB4 is attenuator 'DIV.C' output
    SET_BIT(DDRC, 5);             // PC4 is attenuator 'DIV.B' output
    SET_BIT(DDRC, 4);             // PC5 is attenuator 'DIV.A' output
    
    SET_BIT(DDRD, 1);             // PD1 is test-point output 
    SET_BIT(DDRD, 5);             // PD5 is PWM output 
    SET_BIT(DDRD, 6);             // PD6 is reserved (unused)
    SET_BIT(PORTD, 7);            // PD7 pull-up enabled for ALT button

    outputMode = MODE_SINE;       // startup output mode
    oscFreqRange = HIGH_RANGE;    // startup wave osc freq range
    oscFreq_Hz = 1000;            // startup wave osc freq.
    freqPresetHigh = 10;          // startup wave osc freq preset (high range)
    freqPresetLow = 9;            // startup wave osc freq preset (low range)
    
    pulseFreqRange = HIGH_RANGE;  // startup pulse freq range
    pulseFreq_Hz = 20;            // startup pulse frequency
    pulsePresetHigh = 6;          // startup pulse freq preset (high range)
    pulsePresetLow = 6;           // startup pulse freq preset (low range)
    
    noiseFilterIdx = 0;           // startup noise preset (0 -> white noise)
    outputLevel = 3;              // startup o/p level (3 -> 1Vpp)
    AdjustOutputLevel(NO_CHANGE);
    
    pulseDuty_fxt = WAVE_TABLE_SIZE << 15;     // initial duty (50%)
    freqSettingMode = SET_FREQ_USING_BUTTONS;  // startup freq setting mode
    lastPotSetting = ADC_ReadInput(3) / 4;

    // Calculate the oscillator 'phase step' value using the startup osc. frequency...
    oscStep = ((float) oscFreq_Hz * (WAVE_TABLE_SIZE << 16)) / SAMPLE_RATE; 
    
    // Copy default wave-table from flash program memory
    PGM_ReadData(sineWave, waveTable, WAVE_TABLE_SIZE);

    TC1_TimerInit();              // Set up TC1 for 50ms time interval
    TC0_Setup(5, 1, 250, 125);    // Set up TC0 for PWM output 
    TC0_OCA_IRQ_ENABLE();         // Enable IRQ for PWM output
    GLOBAL_INT_ENABLE();

    lcd_initialise();             // Initialize LCD module
    lcd_command(LCD_CURSOR_OFF);  // Display ON, Cursor OFF
    
    lcd_cursor_posn(0, 0);        // Display start-up message
    lcd_print_string(" Nano Wave-Gen");
    lcd_cursor_posn(1, 0);
    lcd_print_string(" Firmware v");
    lcd_print_string(FIRMWARE_VERSION);
    
    while (startupTimer < 40)     // loop for 2 seconds (= 40 x 50ms)
    {
        if (TimerExpired_50ms())  startupTimer++;
    }        

    while (TRUE)  // loop forever
    {
        if (TimerExpired_50ms())  // Every 50ms, do these tasks...
        {
            ButtonScan(3);  // Detect button hits
            
            potSetting = ADC_ReadInput(3) / 4;  // Update current pot setting
            
            if (outputMode != MODE_NOISE && outputMode != MODE_PULSE && oscFreqRange == HIGH_RANGE)
            {
                if (abs((int)potSetting - (int)lastPotSetting) > 16)  // knob has moved!
                {
                    freqSettingMode = SET_FREQ_USING_POT;
                    lastPotSetting = potSetting;
                }
            }
            
            if (freqSettingMode == SET_FREQ_USING_POT) SetOscFreqUsingPot();
            if (outputMode == MODE_PULSE)  SetPulseDutyUsingPot(); 
            if (outputMode == MODE_NOISE)  SetNoiseFreqUsingPot(); 

            if (++count_to_4 >= 4)  // Every 200ms, update display
            {
                count_to_4 = 0;
                DisplayUpdate();
            }
        }
        
        if (button_hit('A'))  // MODE/RANGE button hit...
        {
            if (ALT_BUTTON_IS_PRESSED)  // Toggle freq. range
            {
                if (outputMode == MODE_PULSE) pulseFreqRange ^= 1;
                else  oscFreqRange ^= 1; 
            }
            else   // Scroll to next mode
            {
                if (++outputMode >= MODE_UNDEFINED)  
				{
					outputMode = MODE_SINE;
					freqSettingMode = SET_FREQ_USING_BUTTONS;
					lastPotSetting = ADC_ReadInput(3) / 4;
				}
            }

            ConfigureOutputMode();
            AdjustPulseFreqPreset(NO_CHANGE);   // Restore pulse freq. preset
            AdjustOscFreqPreset(NO_CHANGE);     // Restore wave osc freq. preset
        }

        if (button_hit('B'))  // Decrement (-) button hit
        {
            if (outputMode == MODE_NOISE)
            {
                if (ALT_BUTTON_IS_PRESSED) AdjustOutputLevel(STEP_DOWN);
            }
            else if (outputMode == MODE_PULSE) 
            {
                if (ALT_BUTTON_IS_PRESSED && pulseFreqRange == LOW_RANGE)
                      AdjustOutputLevel(STEP_DOWN);
                else  AdjustPulseFreqPreset(STEP_DOWN); 
            }
            else  // Wave-table oscillator modes
            {
                if (ALT_BUTTON_IS_PRESSED) AdjustOutputLevel(STEP_DOWN);
                else  AdjustOscFreqPreset(STEP_DOWN);
            }
        }
        
        if (button_hit('C'))  // Increment (+) button hit
        {
            if (outputMode == MODE_NOISE)
            {
                if (ALT_BUTTON_IS_PRESSED) AdjustOutputLevel(STEP_UP);
            }
            else if (outputMode == MODE_PULSE)
            {
                if (ALT_BUTTON_IS_PRESSED && pulseFreqRange == LOW_RANGE) 
                      AdjustOutputLevel(STEP_UP);
                else  AdjustPulseFreqPreset(STEP_UP);
            }
            else  // Wave-table oscillator modes
            {
                if (ALT_BUTTON_IS_PRESSED) AdjustOutputLevel(STEP_UP);
                else  AdjustOscFreqPreset(STEP_UP);
            }		
        }
    }
}


void  ConfigureOutputMode()
{
    const  BYTE  *wavdat = sineWave;  // pointer to wave-table data
    
    if (outputMode != MODE_NOISE && outputMode != MODE_PULSE)
    {
        if (outputMode == MODE_SINE) wavdat = sineWave;
        else if (outputMode == MODE_TRIANGLE) wavdat = triangleWave;
        else  wavdat = sawtoothWave;   // May be POS or NEG slope
        
        // Copy selected wave-table from flash program memory
        PGM_ReadData(wavdat, waveTable, WAVE_TABLE_SIZE);
        
        if (outputMode == MODE_SAWTOOTH_NEG)  // NEG slope
            ReverseOrder((BYTE *) waveTable, WAVE_TABLE_SIZE); 
    }
    
    if (outputMode == MODE_PULSE && pulseFreqRange == HIGH_RANGE)
    {
        TC0_OCA_IRQ_DISABLE();      // High freq PULSE mode -- no ISR
        TC0_Setup(7, 1, 250, 125);  // Initialize TC0 for Fast PWM
    }
    else  // Wave-table sampling modes
    {
        TC0_Setup(5, 1, 250, 125);  // Set up TC0 for Phase-correct PWM
        TC0_OCA_IRQ_ENABLE();       // Sampling modes use PWM ISR
    }
    
	if (outputMode == MODE_PULSE && pulseFreqRange == LOW_RANGE)
	    SET_BIT(PORTB, 5);          // Enable PWM filter (LRANGE = 1)
	else  CLEAR_BIT(PORTB, 5);      // Disable PWM filter (LRANGE = 0)
	
    if (outputMode == MODE_NOISE || outputMode == MODE_PULSE || oscFreqRange == LOW_RANGE)
        freqSettingMode = SET_FREQ_USING_BUTTONS;  // forget pot setting
}


void  AdjustOscFreqPreset(BYTE direction)
{
    // Table of 18 frequency preset values -- for wave-table osc. High range
    static unsigned highFreqTable[] =
        { 80, 100, 160, 200, 250, 320, 400, 500, 640, 800, 1000, 1600, 2000, 3200, 4000, 5000, 6400, 8000 };
    // Table of 15 frequency preset values -- for wave-table osc. Low range
    static BYTE lowFreqTable[] =
        { 1, 2, 4, 5, 8, 10, 16, 20, 25, 32, 40, 50, 64, 80, 100 };
            
    freqSettingMode = SET_FREQ_USING_BUTTONS;  // until knob moved again

    if (oscFreqRange == HIGH_RANGE)
    {
        if (direction == STEP_UP && freqPresetHigh < 17)  freqPresetHigh++;
        if (direction == STEP_DOWN && freqPresetHigh > 0)  freqPresetHigh--;

        oscFreq_Hz = highFreqTable[freqPresetHigh];
    }
    else  // oscFreqRange == LOW_RANGE
    {
        if (direction == STEP_UP && freqPresetLow < 14)  freqPresetLow++;
        if (direction == STEP_DOWN && freqPresetLow > 0)  freqPresetLow--;

        oscFreq_Hz = lowFreqTable[freqPresetLow];
    }

    // Set the oscillator "phase angle step" according to oscFreq_Hz
    oscStep = ((float) oscFreq_Hz * (WAVE_TABLE_SIZE << 16)) / SAMPLE_RATE; 
}


void  AdjustPulseFreqPreset(BYTE direction)
{
    // Table of 16 frequency preset values -- for Pulse generator High range
    static unsigned long highFreqTable[] =
        { 1000, 2000, 2500, 5000, 10000, 20000, 25000, 40000, 50000, 
          100000, 200000, 250000, 500000, 1000000, 2000000, 4000000 };
    // Table of 16 frequency preset values -- for Pulse generator Low range
    static unsigned lowFreqTable[] =
        { 1, 2, 4, 5, 8, 10, 20, 40, 50, 80, 100, 200, 400, 500, 800, 1000 };

    if (pulseFreqRange == HIGH_RANGE)
    {
        if (direction == STEP_UP && pulsePresetHigh < 15)  pulsePresetHigh++;
        if (direction == STEP_DOWN && pulsePresetHigh > 0)  pulsePresetHigh--;

        pulseFreq_Hz = highFreqTable[pulsePresetHigh];
    }
    else  // pulseFreqRange == LOW_RANGE
    {
        if (direction == STEP_UP && pulsePresetLow < 15)  pulsePresetLow++;
        if (direction == STEP_DOWN && pulsePresetLow > 0)  pulsePresetLow--;

        pulseFreq_Hz = lowFreqTable[pulsePresetLow];  
        oscStep = ((float) pulseFreq_Hz * (WAVE_TABLE_SIZE << 16)) / SAMPLE_RATE;
    }
}


void  AdjustOutputLevel(BYTE dirn)
{
    if (dirn == STEP_UP && outputLevel < 5)  outputLevel++;  // outputLevel = 0..5
    if (dirn == STEP_DOWN && outputLevel > 0)  outputLevel--;
	
	if (outputLevel > 2) SET_BIT(PORTB, 4);  // DIV.C = 1
	else  CLEAR_BIT(PORTB, 4);  // DIV.C = 0
	
	CLEAR_BIT(PORTC, 4);  // DIV.B = 0
	CLEAR_BIT(PORTC, 5);  // DIV.A = 0
    
	if (outputLevel == 0 || outputLevel == 3)  SET_BIT(PORTC, 4);  // DIV.B = 1
    if (outputLevel == 1 || outputLevel == 4)  SET_BIT(PORTC, 5);  // DIV.A = 1
}


/*
 * The current pot setting is used to set the oscillator frequency.
 * Apply square-law transform to obtain approx exponential freq vs pot curve.
 * This function is called when the output mode is not NOISE or PULSE.
 */
void  SetOscFreqUsingPot()
{
    unsigned  adc_count = ADC_ReadInput(3);   // 0..1023
    long  potSquared = ((long) adc_count * adc_count) / 1024;
    
    oscFreq_Hz = 5 * potSquared;               // 0..5115
    if (oscFreq_Hz < 50) oscFreq_Hz = 50;      // min. 50
    if (oscFreq_Hz > 5000) oscFreq_Hz = 5000;  // max. 5000

    // Set the oscillator "phase angle step" size according to oscFreq_Hz
    oscStep = ((float) oscFreq_Hz * (WAVE_TABLE_SIZE << 16)) / SAMPLE_RATE;
}

/*
 * The current pot setting is used to set the Noise Filter time-constant.
 * Actually, the variable noiseFilterIdx is the log (base 2) of the coeff (1/K),
 * so the actual coefficient values are K = 1, 2, 4, 8, 16 ... 128. 
 * See DisplayUpdate() function for corresponding corner frequencies in Hz.
 * A value of 0 for noiseFilterIdx disables the filter, producing "white" noise.
 * This function is called only when output mode is NOISE.
 */
void  SetNoiseFreqUsingPot()
{
    noiseFilterIdx = 7 - potSetting / 32;  // Pot quantized into 8 steps (0..7)
}


/*
 * The current pot setting is used to set the pulse duty for the selected range.
 * The duty value is quantized according to a table of preset percentages.
 */
void  SetPulseDutyUsingPot()
{
    // Table of 16 preset pulse duty values
    static BYTE pulseDutyTable[] =
        { 1, 2, 5, 10, 15, 20, 25, 50, 50, 75, 80, 85, 90, 95, 98, 99 };
            
    // Timer period (number of TC0 clock pulses) corresponding to pulsePresetHigh (0..15)
    static BYTE pulsePeriodTable[] = 
        { 250, 125, 100, 50, 200, 100, 80, 50, 40, 160, 80, 64, 32, 16, 8, 4 };
    
    BYTE  PS_bits = 3;        // TC0 Prescaler bits setting in TCSR0B
    BYTE  duty_clks = 1;      // Timer duty (number of TC0 clock pulses)
    BYTE  period_clks = pulsePeriodTable[pulsePresetHigh];  // High freq range only
    
    if (pulseFreqRange == HIGH_RANGE)
    {
        pulseDuty_pc = pulseDutyTable[potSetting/16];  
        duty_clks = ((long) period_clks * pulseDuty_pc) / 100;
        
        // Now find the actual duty (percent, nearest integer) and pulse width (ns)
        if (pulsePresetHigh <= 3)   // pulse Freq <= 5kHz ...
        {
            PS_bits = 3;    // Prescaler N = 64
            pulseDuty_pc = ((int)duty_clks * 100) / period_clks;
            pulseWidth_ns = (long)duty_clks * 4000;  // Tclk = 4us
        }
        else if (pulsePresetHigh <= 8)  //  pulse Freq <= 50kHz ...
        {
            PS_bits = 2;    // Prescaler N = 8
            pulseDuty_pc = ((int)duty_clks * 100) / period_clks;
            pulseWidth_ns = (long)duty_clks * 500;  // Tclk = 500ns
        }
        else  //  pulse Freq > 50kHz ...
        {
            PS_bits = 1;    // Prescaler N = 1
            pulseDuty_pc = ((int)duty_clks * 100) / period_clks;
            pulseWidth_ns = ((int)duty_clks * 125) / 2;  // Tclk = 62.5ns
        }
        
        TC0_WRITE_PRESCALER(PS_bits);
        OCR0A = period_clks - 1;
        OCR0B = duty_clks;
    }
    else   // pulseFreqRange == LOW_RANGE
    {
        pulseDuty_pc = pulseDutyTable[potSetting/16];
        
        // Convert percent duty to fixed-point value (integer part)
        pulseDuty_fxt = ((WAVE_TABLE_SIZE << 16) * pulseDuty_pc) / 100;
    }
}


/*
 * Function:  DisplayUpdate()
 *
 * This function updates information displayed on the LCD.
 * Called at least 5 times per second.
 */
void  DisplayUpdate()
{
    static BYTE lastOutputMode = 99;  // invoke refresh
    static char *waveformName[] = 
        { "SINEWAVE", "TRIANGLE", "SQUARE", "SAWTOOTH+", "SAWTOOTH-", "PULSE", "NOISE" };
    static char *levelText[] = { "100mV", "200mV", "500mV", " 1Vpp", " 2Vpp", " 4Vpp" };
    static char *noiseFreqText[] =  
        { "----", "3200", "1600", " 800", " 400", " 200", " 100", "  50" };
        
    char   strBuff[16];
    char   curpos = 0;
    unsigned long  period_us = 0;   // TBD
    unsigned int   pulseWidth = 0;  // TBD
    
    if (ALT_BUTTON_IS_PRESSED)
    {
        lcd_erase_field(0, 0, 16);    // Info on top line...
        
        if (outputMode == MODE_NOISE)  lcd_print_string("NOISE");
        else if (outputMode == MODE_PULSE)
        {
            if (pulseFreqRange == HIGH_RANGE)
            lcd_print_string("RANGE: 1k..4M Hz");
            else  lcd_print_string("RANGE:  1..1k Hz");
        }
        else  // Wave-table sampling modes...
        {
            if (oscFreqRange == LOW_RANGE)
                lcd_print_string("RANGE: 1..100 Hz");
            else  lcd_print_string("RANGE: 80..8k Hz");  
        }
        
        lcd_erase_field(1, 0, 16);    // Info on bottom line...
        
        if (outputMode != MODE_PULSE || pulseFreqRange == LOW_RANGE)
        {
            lcd_print_string("LEVEL: ");
            lcd_print_string(levelText[outputLevel]);
            if (outputMode == MODE_PULSE) lcd_print_string(" AC");
			else  lcd_print_string("   ");
        }
        
        // Display a symbol (*) if osc freq setting mode is using the pot.
        lcd_cursor_posn(1, 15);
        if (freqSettingMode == SET_FREQ_USING_POT)  lcd_write_char('*');

        lastOutputMode = 99;  // Force display refresh when ALT button released
        
        return;   // All done  (if ALT button pressed)
    }
    
    // If output mode has changed, refresh fixed data on top line
    if (outputMode != lastOutputMode)  
    {
        lcd_erase_field(0, 0, 16); 
        if (outputMode != MODE_PULSE) lcd_print_string("Mode: ");
        lcd_print_string(waveformName[outputMode]);
        lastOutputMode = outputMode;
    }

    if (outputMode == MODE_NOISE)
    {
        lcd_cursor_posn(1, 0);
        // Display noise filter corner freq in Hz on bottom line
        if (noiseFilterIdx != 0)
        {
            lcd_print_string("Fc:");
            lcd_print_string(noiseFreqText[noiseFilterIdx]);
            lcd_print_string("Hz ");
        }
        else  lcd_print_string("No filter ");
        
        // Display output level setting, also on bottom line
        lcd_cursor_posn(1, 10);
        lcd_print_string(levelText[outputLevel]);
		lcd_print_string("    ");
    }
    else if (outputMode == MODE_PULSE)
    {
        // Display pulse frequency on bottom line LHS 
        unsigned  freqVal = pulseFreq_Hz;  
        lcd_cursor_posn(1, 0);
        lcd_print_string("Fo:    Hz  ");
        
        if (pulseFreq_Hz >= 1000000)  // display in MHz
        {
            freqVal = pulseFreq_Hz / 1000000;
            lcd_cursor_posn(1, 6);
            lcd_write_char('M');
        }
        else if (pulseFreq_Hz >= 1000)  // display in kHz
        {
            freqVal = pulseFreq_Hz / 1000;
            lcd_cursor_posn(1, 6);
            lcd_write_char('k');
        }
        itoa(freqVal, strBuff, 10);
        curpos = 6 - strlen(strBuff);
        if (pulseFreq_Hz < 1000) curpos++;
        lcd_cursor_posn(1, curpos);  lcd_print_string(strBuff);

        // Display pulse duty in percent on bottom line RHS
		if (pulseDuty_pc == 0)  pulseDuty_pc = 1;
        lcd_cursor_posn(1, 11);  lcd_print_string("d:  %");
        itoa(pulseDuty_pc, strBuff, 10);
        curpos = 15 - strlen(strBuff);
        lcd_cursor_posn(1, curpos);  lcd_print_string(strBuff);
        
        // Display pulse width (Tpw) on top line  
        lcd_cursor_posn(0, 7);  lcd_print_string("Tpw:   us");
        
        if (pulseFreqRange == LOW_RANGE)
        {
            period_us = 1000000UL / pulseFreq_Hz;
            pulseWidth = (period_us * pulseDuty_pc) / 100;  // us
            if (pulseWidth >= 1000)  // us
            {
                pulseWidth /= 1000;   // pulseWidth unit is 1ms
                lcd_cursor_posn(0, 14);  lcd_write_char('m');
            }
        }
        else  // pulseFreqRange == HIGH_RANGE
        {
            if (pulseWidth_ns >= 1000)  pulseWidth = pulseWidth_ns / 1000;  // us
            else  
            {
                pulseWidth = pulseWidth_ns;   // pulseWidth < 1000 ns
				if (pulseWidth == 0)  pulseWidth = 1;
                lcd_cursor_posn(0, 14);  lcd_write_char('n');
            }
        }
        itoa(pulseWidth, strBuff, 10);
        curpos = 14 - strlen(strBuff);
        lcd_cursor_posn(0, curpos);  lcd_print_string(strBuff);
    }
    else  // Wave-table Oscillator modes
    {
        // Display osc frequency in Hz (4 digits)
        itoa(oscFreq_Hz, strBuff, 10);
        curpos = 7 - strlen(strBuff);
        lcd_cursor_posn(1, 0);
        lcd_print_string("Fo:    Hz   ");
        lcd_cursor_posn(1, curpos); 
        lcd_print_string(strBuff);
        
        // Display output level setting
        lcd_cursor_posn(1, 10);
        lcd_print_string(levelText[outputLevel]);
		lcd_print_string("  ");
    }
}


/*
 * Function:  TC0_Setup()
 *
 * This function initializes Timer-Counter TC0 in Fast PWM mode or Phase-correct
 * PWM mode to generate a variable-duty pulse waveform on pin PD5/OC0B.
 *
 * Entry arg's:     wgm = waveform gen. mode (5: Phase-Correct PWM, 7: Fast PWM)
 *                presc = Prescaler code (1..5 : N = 1, 8, 64, 256, 1024, resp.)
 *               period = value written to OCR0A reg. (minus 1)
 *                 duty = value written to OCR0B reg.
 *
 * TC0 setup example for phase-correct PWM mode (WGM = 5):
 * ```````````````````````````````````````````````````````
 * With prescaler = 1, period (OCR0A reg.) is set to 249 for a complete up/down
 * timer cycle of 500 clocks, so the PWM output freq. will be 16MHz / 500 = 32kHz.
 */
void  TC0_Setup(BYTE wgm, BYTE presc, BYTE period, BYTE duty)
{
    TCCR0A = 0;   // Disable timer/counter during setup
    TCCR0B = 0;
    OCR0A = period - 1;
    OCR0B = duty;
    
    TCCR0A = 0b00100000 | (wgm & 0b11);   // Write WGM b1, b0
    TCCR0B = (presc & 0b111) | ((wgm << 1) & 0b1000);  // Write WGM b2
}


/*
 * Timer-Counter TC0 Interrupt Service Routine
 *
 * The ISR is called at the waveform generation sample rate (32kHz).
 * The PWM duty (set by register OCR0B) is updated with the next sample.
 *
 * The high-freq PULSE mode does not use a sampling method, hence the Timer IRQ
 * is disabled for this mode. The low-freq pulse mode uses a sampling algorithm,
 * but without a predefined wave-table. Sample values are computed "on-the-fly".
 * Likewise, the square-wave mode computes samples without a stored wave-table.
 *
 * A detailed description of the wave-table oscillator algorithm is given in a
 * separate document (see References).
 */
ISR( TIMER0_COMPA_vect )
{
    static unsigned long  oscAngle;       // fixed-point (32 bits)
    static unsigned long  rand_last = 1;  // random number seed
    static unsigned int   yn_last;        // noise filter output sample
    
    SET_BIT(PORTD, 1);       // Test-point PD1 = High

    BYTE  sample = 0;        // PWM duty, range: 0 .. 249
    int   index;             // wave-table index
    unsigned int  rand16;    // 16-bit pseudo-random number
    unsigned int  xn, yn;    // noise filter input, output (8:8 fixed-point)

    if (outputMode == MODE_PULSE)  // assume pulseFreqRange == LOW_RANGE
    {
        if (oscAngle < pulseDuty_fxt) sample = 245;  // High state
        else  sample = 5;   // Low state
        oscAngle = oscAngle + oscStep;
        if (oscAngle >= (WAVE_TABLE_SIZE << 16))  // past end of table
            oscAngle -= (WAVE_TABLE_SIZE << 16);  // wrap around
    }
    else if (outputMode == MODE_SQUARE)  
    {
        // Same algorithm as pulse mode, but duty is fixed at 50% of period
        if (oscAngle < (WAVE_TABLE_SIZE << 15)) sample = 245;  // High state
        else  sample = 5;   // Low state
        oscAngle = oscAngle + oscStep;
        if (oscAngle >= (WAVE_TABLE_SIZE << 16))  // past full cycle (> 360 degrees)
            oscAngle -= (WAVE_TABLE_SIZE << 16);  // wrap around
    }
    else if (outputMode == MODE_NOISE)
    {
        // Get 8-bit random sample value (xn) in the range 5..245
        rand_last = rand_last * 1103515245 + 12345;
        rand16 = rand_last & 0xFFFF;
        xn = 5 + (rand16 % 240);
 
        if (noiseFilterIdx != 0)  // Apply 1st-order IIR filter 
        {
            xn = xn << 8;  // convert noise sample to 16-bit fixed-point
            yn = yn_last - (yn_last >> noiseFilterIdx) + (xn >> noiseFilterIdx);
            sample = yn >> 8;   // Integer part (8 MS bits)
            yn_last = yn;
        }
        else  sample = xn;   // Bypass filter
    }
    else  // Wave-table oscillator modes (sine, triangle, sawtooth)
    {
        index = oscAngle >> 16;     // index = integer part of oscAngle
        sample = waveTable[index];
        oscAngle = oscAngle + oscStep;
        if (oscAngle >= (WAVE_TABLE_SIZE << 16))  // past end of table
            oscAngle -= (WAVE_TABLE_SIZE << 16);  // wrap around
    }

    OCR0B = sample;   // update PWM duty with sample value
    
    CLEAR_BIT(PORTD, 1);    // Test-point PD1 = Low
}


/*
 * Timer TC1 is set up to create a recurring 50ms event timer.
 * The library function 'milliseconds()' cannot be used, because it uses an
 * interrupt which would interfere with the wave-table oscillator ISR.
 *
 * Set up timer TC1 in Normal mode -- count up, rollover, no interrupt.
 * Prescaler:  F_CPU / 64  -->  Tclk = 4us (@ F_CPU = 16MHz)
 * Count == 12500 -> 50,000us -> 50ms
 */
void  TC1_TimerInit()
{
    TCCR1B = 0;            // Disable TC1 during setup
    TCCR1A = 0;            // Normal mode
    TCNT1 = 0;             // Reset timer count reg.
    TCCR1B = 0b00000011;   // Enable timer;  Pre-scaler = 1/64;
}


/*
 * TC1 Timer "service routine" -- Polling mode (non-interrupt).
 * This function must be called *frequently* from the main loop.
 *
 * The timer count register is checked;  if it is > 50ms, i.e. timer expired,
 * then the timer is reset (count = 0) and the function returns TRUE.
 * Otherwise, it returns FALSE.
 */
BOOL  TimerExpired_50ms()
{
    BOOL  result = FALSE;
    unsigned  count = TCNT1;

    if (count >= 12500)
    {
        result = TRUE;
        TCNT1 = 0;
    }

    return  result;
}


/*
 * Reverse order of data in an array of a given size (bytes).
 */
void  ReverseOrder(BYTE *dest, unsigned size)
{
    BYTE  b;
    BYTE  *src = dest + (size - 1);  // address of last byte
    unsigned swaps = size / 2;  
	
    while (swaps-- != 0)
    {
        b = *dest;
        *dest++ = *src;
        *src-- = b;
    }
}


/*
 * Write a number of blank spaces (nplaces) to LCD starting at (row, col).
 * The function restores the cursor position to (row, col) on exit.
 */
void  lcd_erase_field(BYTE row, BYTE col, BYTE nplaces)
{
    lcd_cursor_posn(row, col);
    
    while (nplaces--)  { lcd_write_char(' '); }
        
    lcd_cursor_posn(row, col);
}


/* ````````````````````````````````````````````````````````````````````````````
 * Function:  PGM_ReadData()
 *
 * Copy data from flash program memory to an array in data memory (SRAM).
 *
 * Entry args:  src    =  address of source data in program memory
 *              dest   =  address of destination array in data memory
 *              nbytes =  number of bytes to be copied
 * 
 * The macro pgm_read_byte() is defined in header file avr/pgmspace.h
 * which must be #included in source files which call this function.
 *
 * Flash memory data must be declared as in this example:
 * const char array_name[] PROGMEM = { k0, k1, k2, ... kn };
 */
void  PGM_ReadData(const BYTE *src, BYTE *dest, unsigned nbytes)
{
    while (nbytes != 0)
    {
        *dest++ = pgm_read_byte(src++);
        --nbytes;
    }
}

// end-of-file
